@@ -19,6 +19,7 @@
  */
 
 #include "config.h"
+#include <locale.h>
 #include <libgnome/libgnome.h>
 #include <libgnomeui/libgnomeui.h>
 #include <string.h>
@@ -29,7 +30,11 @@ typedef struct _Language Language;
 struct _Language {
 	char *name;
 	char *code;
+	char *untranslated;
 	int found;
+
+	/* extra fields */
+	char *collate_key;
 };
 
 /* Note: these should NOT include the encodings, this is just a translation
@@ -37,78 +42,91 @@ struct _Language {
  * languages, just their names and where they are placed in the menu.
  * The available languages come from the supplied locale.alias */
 static Language languages [] = {
+/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Azerbaijani"), "az_AZ", NULL, 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Catalan"), "ca_ES", 0 },
+	{ N_("A-M|Basque"), "eu_ES", NULL, 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Chinese (simplified)"), "zh_CN", 0 },
+	{ N_("A-M|Bulgarian"), "bg_BG", NULL, 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Chinese (traditional)"), "zh_TW", 0 },
+	{ N_("A-M|Catalan"), "ca_ES", "Català", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Croatian"), "hr_HR", 0 },
+	{ N_("A-M|Chinese (simplified)"), "zh_CN", "中文 (简体)", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Czech"), "cs_CZ", 0 },
+	{ N_("A-M|Chinese (traditional)"), "zh_TW", "繁體中文", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Danish"), "da_DK", 0 },
+	{ N_("A-M|Croatian"), "hr_HR", "Hrvatski", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Dutch"), "nl_NL", 0 },
+	{ N_("A-M|Czech"), "cs_CZ", "čeština", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|English"), "en", 0 },
+	{ N_("A-M|Danish"), "da_DK", "dansk", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|American English"), "en_US", 0 },
+	{ N_("A-M|Dutch"), "nl_NL", "Nederlands", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|British English"), "en_GB", 0 },
+	{ N_("A-M|English"), "en", "English", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Finnish"), "fi_FI", 0 },
+	{ N_("A-M|Estonian"), "et_EE", "Eesti", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|French"), "fr_FR", 0 },
+	{ N_("A-M|American English"), "en_US", "American English", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Galician"), "gl_ES", 0 },
+	{ N_("A-M|British English"), "en_GB", "British English", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|German"), "de_DE", 0 },
+	{ N_("A-M|Finnish"), "fi_FI", "Suomi", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Greek"), "el_GR", 0 },
+	{ N_("A-M|French"), "fr_FR", "Français", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Hebrew"), "iw_IL", 0 },
+	{ N_("A-M|Galician"), "gl_ES", "Galego", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Hungarian"), "hu_HU", 0 },
+	{ N_("A-M|German"), "de_DE", "Deutsch", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Icelandic"), "is_IS", 0 },
+	{ N_("A-M|Greek"), "el_GR", "ελληνικά", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Italian"), "it_IT", 0 },
+	{ N_("A-M|Hebrew"), "he_IL", "עברית", 0 },
+	{ N_("A-M|Hebrew"), "iw_IL", "עברית", 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Japanese"), "ja_JP", 0 },
+	{ N_("A-M|Hungarian"), "hu_HU", NULL, 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Korean"), "ko_KR", 0 },
+	{ N_("A-M|Icelandic"), "is_IS", NULL, 0 },
 	/*Note translate the A-M to the A-M you used in the group label */
-	{ N_("A-M|Lithuanian"), "lt_LT", 0 },
+	{ N_("A-M|Italian"), "it_IT", "Italiano", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Japanese"), "ja_JP", "日本語", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Korean"), "ko_KR", "한국어", 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Latvian"), "lv_LV", NULL, 0 },
+	/*Note translate the A-M to the A-M you used in the group label */
+	{ N_("A-M|Lithuanian"), "lt_LT", "Lietuvių", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Norwegian (bokmal)"), "no_NO", 0 },
+	{ N_("N-Z|Norwegian (bokmal)"), "no_NO", NULL, 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Norwegian (nynorsk)"), "nn_NO", 0 },
+	{ N_("N-Z|Norwegian (nynorsk)"), "nn_NO", NULL, 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Polish"), "pl_PL", 0 },
+	{ N_("N-Z|Polish"), "pl_PL", "Polski", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Portuguese"), "pt_PT", 0 },
+	{ N_("N-Z|Portuguese"), "pt_PT", "Português", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Brazilian Portuguese"), "pt_BR", 0 },
+	{ N_("N-Z|Brazilian Portuguese"), "pt_BR", "Português do Brasil", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Romanian"), "ro_RO", 0 },
+	{ N_("N-Z|Romanian"), "ro_RO", "Română", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Russian"), "ru_RU", 0 },
+	{ N_("N-Z|Russian"), "ru_RU", "русский", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Slovak"), "sk_SK", 0 },
+	{ N_("N-Z|Slovak"), "sk_SK", "Slovenský", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Slovenian"), "sl_SI", 0 },
+	{ N_("N-Z|Slovenian"), "sl_SI", "Slovenščina", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Spanish"), "es_ES", 0 },
+	{ N_("N-Z|Spanish"), "es_ES", "Español", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Swedish"), "sv_SE", 0 },
+	{ N_("N-Z|Swedish"), "sv_SE", "Svenska", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Turkish"), "tr_TR", 0 },
+	{ N_("N-Z|Turkish"), "tr_TR", "Türkçe", 0 },
 	/*Note translate the N-Z to the N-Z you used in the group label */
-	{ N_("N-Z|Ukrainian"), "uk_UA", 0 },
+	{ N_("N-Z|Ukrainian"), "uk_UA", NULL, 0 },
+	/*Note translate the N-Z to the N-Z you used in the group label */
+	{ N_("N-Z|Walloon"), "wa_BE", NULL, 0 },
 	/* This is the POSIX/C locale for english, should really be in Other */
-	{ N_("Other|POSIX/C English"), "C", 0 },
+	{ N_("Other|POSIX/C English"), "C", NULL, 0 },
 	{ NULL, NULL }
 };
 
@@ -181,19 +199,17 @@ find_lang (const char *language, gboolean *clean)
 }
 
 char *
-gdm_lang_name (const char *locale, const char *language,
-	       gboolean never_encoding, gboolean no_group)
+gdm_lang_name (const char *language,
+	       gboolean never_encoding,
+	       gboolean no_group,
+	       gboolean untranslated,
+	       gboolean markup)
 {
 	Language *lang;
 	char *name;
 	gboolean clean;
 	const char *encoding;
 
-	/* FIXME: we need to get the languages for right locales here,
-	 * for now we just translate the current locale */
-	if (locale != NULL)
-		return NULL;
-
 	gdm_lang_init ();
 
 	lang = find_lang (language, &clean);
@@ -213,6 +229,22 @@ gdm_lang_name (const char *locale, const char *language,
 	else
 		name = g_strdup (_(lang->name));
 
+	if (lang->untranslated != NULL && untranslated) {
+		char *full;
+		if (markup) {
+			full = g_strdup_printf
+				("%s (<span lang=\"%s\">%s</span>)",
+				 name,
+				 lang->code,
+				 lang->untranslated);
+		} else {
+			full = g_strdup_printf ("%s (%s)",
+						name, lang->untranslated);
+		}
+		g_free (name);
+		name = full;
+	}
+
 	if (no_group) {
 		char *p = strchr (name, '|');
 		if (p != NULL) {
@@ -223,11 +255,27 @@ gdm_lang_name (const char *locale, const char *language,
 	return name;
 }
 
-GdkFont *
-gdm_lang_font (const char *locale)
+/* NULL if not found */
+char *
+gdm_lang_untranslated_name (const char *language,
+			    gboolean markup)
 {
-	/* FIXME! */
-	return NULL;
+	Language *lang;
+	gboolean clean;
+
+	gdm_lang_init ();
+
+	lang = find_lang (language, &clean);
+	if (lang == NULL)
+		return NULL;
+
+	if (markup && lang->untranslated != NULL) {
+		return g_strdup_printf ("<span lang=\"%s\">%s</span>",
+					lang->code,
+					lang->untranslated);
+	} else {
+		return g_strdup (lang->untranslated);
+	}
 }
 
 const char *
@@ -252,24 +300,32 @@ lang_collate (gconstpointer a, gconstpointer b)
 	gboolean clean;
 	Language *l1 = find_lang ((const char *)a, &clean);
 	Language *l2 = find_lang ((const char *)b, &clean);
-	const char *name1, *name2;
 
 	/* paranoia */
 	if (l1 == NULL || l2 == NULL)
 		return 0;
 
-	name1 = strchr (_(l1->name), '|');
-	if (name1 != NULL)
-		name1 ++;
-	else
-		name1 = _(l1->name);
-	name2 = strchr (_(l2->name), '|');
-	if (name2 != NULL)
-		name2 ++;
-	else
-		name2 = _(l2->name);
+	if (l1->collate_key == NULL) {
+		const char *name;
+		name = strchr (_(l1->name), '|');
+		if (name != NULL)
+			name ++;
+		else
+			name = _(l1->name);
+		l1->collate_key = g_utf8_collate_key (name, -1);
+	}
 
-	return strcoll (name1, name2);
+	if (l2->collate_key == NULL) {
+		const char *name;
+		name = strchr (_(l2->name), '|');
+		if (name != NULL)
+			name ++;
+		else
+			name = _(l2->name);
+		l2->collate_key = g_utf8_collate_key (name, -1);
+	}
+
+	return strcmp (l1->collate_key, l2->collate_key);
 }
 
 GList *
@@ -280,6 +336,9 @@ gdm_lang_read_locale_file (const char *locale_file)
 	GList *langs = NULL;
 	GHashTable *dupcheck;
 	gboolean got_C = FALSE;
+	Language *language;
+	gboolean clean;
+	char *curlocale;
 
 	if (locale_file == NULL)
 		return NULL;
@@ -296,8 +355,6 @@ gdm_lang_read_locale_file (const char *locale_file)
 	while (fgets (curline, sizeof (curline), langlist)) {
 		char *name;
 		char *lang;
-		gboolean clean;
-		Language *language;
 
 		if (curline[0] <= ' ' ||
 		    curline[0] == '#')
@@ -323,6 +380,7 @@ gdm_lang_read_locale_file (const char *locale_file)
 			language->found = 1;
 			language->name = g_strdup (name);
 			language->code = g_strdup (lang);
+			language->untranslated = NULL;
 			g_hash_table_insert (lang_names,
 					     language->code,
 					     language);
@@ -342,6 +400,13 @@ gdm_lang_read_locale_file (const char *locale_file)
 	if ( ! got_C)
 		langs = g_list_prepend (langs, g_strdup ("C"));
 
+	curlocale = setlocale (LC_MESSAGES, NULL);
+	if (curlocale != NULL &&
+	    strcmp (curlocale, "C") != 0 &&
+	    find_lang (curlocale, &clean) == NULL) {
+		langs = g_list_prepend (langs, g_strdup (curlocale));
+	}
+
 	langs = g_list_sort (langs, lang_collate);
 
 	fclose (langlist);
